/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999-2003 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __prefab_models_h_
#define __prefab_models_h_
#include "platform_opts.h"
#include <map>
#include "adevs_models.h"
/**
A generic coupled model that can be constructed using
any netExec derived NET_EXEC class whose constructor
accepts a single argument which is the parent network
model.
*/
template <class NET_EXEC> class coupled: public devsn
{
	public:
		coupled():devsn(),nx(this){}
		netExec* getNetExec() { return &nx; }
		~coupled(){}
	private:
		NET_EXEC nx;
};

/**
The staticDigraph class can be used to create
static structure digraph models.  The copy constructor,
clone method, and assignment operator rely on the
component models supporting the clone() method.
*/
class staticDigraph: public devsn 
{
	public:
		staticDigraph():devsn()
		{
			nx = new net_exec(this);
		}
		staticDigraph(const staticDigraph& src):
		devsn(src)
		{
			nx = new net_exec(*(src.nx),this);
		}
		const staticDigraph& operator=(const staticDigraph& src)
		{
			devsn::operator=(src);
			delete nx;
			nx = new net_exec(*(src.nx),this);
			return *this;
		}
		/// Add a model to the network
		void add(devs* model) { nx->addModel(model); }
		/// Couple the src model to the destination model
		void couple(devs* src, port_t srcPort, devs* dst, port_t dstPort)
		{
			nx->addCoupling(src,srcPort,dst,dstPort);
		}
		netExec* getNetExec() { return nx; }
		~staticDigraph(){ delete nx; }
		object* clone() const { return new staticDigraph(*this); }
	private:	
		class net_exec: public netExec
		{
			public:
				net_exec(devs* parent = NULL):netExec(parent){}
				net_exec(const net_exec& src, devs* parent);
				void init();
				void delta_int(){}
				void delta_ext(stime_t e, const adevs_bag<PortValue>& x){}
				void delta_conf(const adevs_bag<PortValue>& x){}
				void output_func(adevs_bag<PortValue>& y){}
				void gc_output(adevs_bag<PortValue>& g){}
				void gc_models(adevs_set<devs*>& g){}
				void getComponents(adevs_set<devs*>& c);
				void addModel(devs* d);
				void addCoupling(devs* src, port_t srcPort, devs* dst, port_t dstPort);
				void route(const PortValue& pv, devs* model, adevs_bag<EventReceiver>& r);
				~net_exec();
			private:
				adevs_set<devs*> models;
				adevs_hashtable<EventReceiver,adevs_bag<EventReceiver>,EventReceiverHashFunc >
					graph;
				net_exec(const net_exec& src){}
				void operator=(const net_exec& src){}
		};
		net_exec* nx;
};

/**
A cell interface for use in the cellSpace model.  All models that will
be added to the cellSpace should inherit this interface.
*/
class cell_interface
{
	public:
		/**
		Input grid, where this cell is at 1,1,1.  To find the input from the
		cell displaced from this cell by +1 or -1, add the displacement value
		to 1 to get the correct input port index.  For example, the cell 
		at -1,-1,0 provides input on port 0,0,1.
		*/
		static const port_t in [3][3][3];
		/// Output port (will be sent to neighboring cells)
		static const port_t out;
};

/**
A 3D cellspace.  This class is meant to be useful for solving PDEs, simulating
next event cells, and other types of models represented as spaces of discrete
interacting points.  It can support a six point (Von Neuman) or 26 point (Moore)
neighborhood.
*/
class cellSpace: public devsn
{
	public:

		/// A coordinate in the cell space
		struct coord_t
		{
			/// Coordinate x, y, and z values
			int x, y, z;
			/// Default constructor
			coord_t(int x = 0, int y = 0, int z = 0):
			x(x),
			y(y),
			z(z)
			{
			}
			/// Copy constructor
			coord_t(const coord_t& src):
			x(src.x),
			y(src.y),
			z(src.z)
			{
			}
			/// Assignment operator
			const coord_t& operator=(const coord_t& src)
			{
				x = src.x;
				y = src.y;
				z = src.z;
				return *this;
			}
		};

		/// Flag indicating a six point neighborhood
		static const int SIX_POINT;
		/// Flag indicating a twenty six point neighborhood
		static const int TWENTY_SIX_POINT;
		/// Creates a 0x0x0 cellspace.
		cellSpace():
		devsn()
		{
			nx = new net_exec(this);
		}
		/// Copy constructor.
		cellSpace(const cellSpace& src):
		devsn(src)
		{
			nx = new net_exec(*(src.nx),this);
		}
		/// Assignment operator
		const cellSpace& operator=(const cellSpace& src)
		{
			devsn::operator=(src);
			delete nx;
			nx = new net_exec(*(src.nx),this);
			return *this;
		}
		/// Create an W x H x D cellspace with NULL model entries in the cells.
		cellSpace(int type, int width, int height = 1, int depth = 1):
		devsn()
		{
			nx = new net_exec(this,width,height,depth,type);
		}
		/// Insert a model at the x,y,z position
		void add(devs* model, int x, int y = 0, int z = 0) { nx->addModel(model,x,y,z); }
		/// Get the network executive for this coupled model
		netExec* getNetExec() { return nx; }
		/// Get the model at location x,y,z
		const devs* getModel(int x, int y = 0, int z = 0) const
		{
			return nx->getModel(x,y,z);
		}
		/// Get the port number assigned to edge coordinate x, y, z
		port_t getPort(int x, int y = 0, int z = 0) const
		{
			return nx->getPort(x,y,z);
		}
		/// Get the coordinate assigned to a port
		coord_t getCoordinates(port_t port) const
		{
			return nx->getCoordinates(port);
		}
		/// Get the width of the cellspace
		int getWidth() const { return nx->getWidth(); }
		/// Get the height of the cellspace
		int getHeight() const { return nx->getHeight(); }
		/// Get the depth of the cellspace
		int getDepth() const { return nx->getDepth(); }
		/// Destructor
		~cellSpace(){ delete nx; }
		/// Returns a copy of this cellSpace
		object* clone() const { return new cellSpace(*this); }
		/// Get a distance number for a cell, used with the cellsim simulator
		int getDistance(devs* model) { return nx->getDistance(model); }
	private:	
		class net_exec: public netExec
		{
			public:
				net_exec(devs* parent = NULL, 
				int width = 0, int height = 0, int depth = 0, int type = SIX_POINT);
				net_exec(const net_exec& src, devs* parent);
				void init();
				void delta_int(){}
				void delta_ext(stime_t e, const adevs_bag<PortValue>& x){}
				void delta_conf(const adevs_bag<PortValue>& x){}
				void output_func(adevs_bag<PortValue>& y){}
				void gc_output(adevs_bag<PortValue>& g){}
				void gc_models(adevs_set<devs*>& g){}
				void getComponents(adevs_set<devs*>& c);
				void addModel(devs* d, int x, int y, int z);
				const devs* getModel(int x, int y, int z) const;
				int getWidth() const { return w; }
				int getHeight() const { return h; }
				int getDepth() const { return d; }
				void route(const PortValue& pv, devs* model, adevs_bag<EventReceiver>& r);
				port_t getPort(int x, int y, int z) const;
				port_t getPort(const devs* model) const;
				coord_t getCoordinates(port_t port) const;
				int getDistance(devs* model);
				~net_exec();
			private:
				int w, h, d, type;
				devs**** space;
				std::map<devs*,coord_t> coord_table;

				devs* getModel(port_t port);
				void internal_coupling(const PortValue& pv, devs* model, 
				adevs_bag<EventReceiver>& r);
				void external_input_coupling(const PortValue& pv, devs* model,
				adevs_bag<EventReceiver>& r);
				void external_output_coupling(const PortValue& pv, devs* model,
				adevs_bag<EventReceiver>& r);

				net_exec(const net_exec& src){}
				void operator=(const net_exec& src){}
		};
		net_exec* nx;
};

#endif
