/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2002 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __xdevssim_h_
#define __xdevssim_h_
#include "platform_opts.h"
#include "adevs_models.h"
#include "pdevs.h"
/**
<p>This simulator class is meant to be used as a component
within another simulation environment (e.g. the HLA). 
It implements a reduced version of the coreSimulatorInterface
from the scalable DEVS interface.  The simplification
made here involves removing timestamps from all calls
except deltFunc(t).  It is assumes that injected inputs
and I/O computations occur at the next deltFunc() and
nextTN() respectively.
*/
class xdevssim 
{
	public:
		xdevssim(devs* model):
		IO_up_to_date(false),
		sim(NULL),
		model(model)
		{}
		/**
		Initialize the simulator prior to simulation start.
		The value t0 is the initial simulation time.
		*/
		virtual void initialize(ADEVS_TIME_TYPE t0 = ADEVS_ZERO);
		/**
		Finalize the simulator at the end of a run.  The simulation engine
		will call no other methods after finalize has been called.
		*/
		virtual void finalize(){}
		/**
		Inject an input to be applied at then next deltFunc(t) call.
		The input is either a single PortValue pair or a bag of PortValue
		pairs. 
		*/
		void inject(const PortValue& x) { input.add(x); }
		void inject(const adevs_bag<PortValue>& x) { input.append(x); }
		/// Get the time of next event (tL + ta())
		double nextTN() const { return sim->timeNext(); }
		/**
		Compute output at time nextTN().  
		*/
		const adevs_bag<PortValue>& computeInputOutput();
		/// Advance the simulation clock and compute state at time t <= nextTN
		void deltFunc(ADEVS_TIME_TYPE t);
		/// Get the model attached to this simulator
		const devs* getModel() const { return model; }
		/**
		Attach a model to this simulator.  The new model type
		must match the original model type.  The old model
		is returned to the caller.  The simulator is assumed
		to be properly initialized and ready to go.  This method
		is meant to support optimistic computing, in which an invalid
		model may need to be replaced with a valid checkpoint.
		*/
		devs* attachModel(devs* model);
		/// Destructor
		virtual ~xdevssim();

	private:
		bool IO_up_to_date;
		basic_pdevs_sim* sim;
		devs* model;
		adevs_bag<PortValue> input;

		xdevssim(const xdevssim& src){}
		void operator=(const xdevssim& src){}

};

#endif
