#include <limits.h>
#include "adevs_threads.h"

// List of jobs to be completed by the thread pool
adevs_bag<adevs_threadpool::job_t> adevs_threadpool::jobs;
// Condition variable controlling access to the job list
cond_var adevs_threadpool::job_cvar;
// Flag used to terminate threads in the pool
bool adevs_threadpool::run_threads = true;
// Number of threads not assigned to jobs
int adevs_threadpool::avail_threads = 0;
// Total number of threads in the pool
int adevs_threadpool::total_threads = 0;
// Limit on the size of the thread pool
int adevs_threadpool::thread_limit = INT_MAX;

void adevs_threadpool::set_thread_limit(int limit)
{
	thread_limit = limit;
}

/**
Thread runs in a inifinte loop, processing jobs as they become
available.
*/
void* run_pthread(void* arg)
{
	/// Increment the available and total counts when this thread starts
	adevs_threadpool::total_threads++;
	adevs_threadpool::avail_threads++;
	/// Look for a job to execute
	while (adevs_threadpool::run_threads)
	{
		/// Wait until a job becomes available
		adevs_threadpool::job_t j(adevs_threadpool::get_job());
		/// Update available count to indicate that we are busy
		adevs_threadpool::avail_threads--;
		/// Execute the job
		j.func(j.arg1,j.arg2);
		/// Indicate to the originator of the job that it has completed
		j.outstanding_sem->up();
		/// Increment the available count since we are ready for more work
		adevs_threadpool::avail_threads++;
	}
	/// If the exit flag is set, update counters and exit
	adevs_threadpool::avail_threads--;
	adevs_threadpool::total_threads--;
	return NULL;
}

/// Create a thread pool
adevs_threadpool::adevs_threadpool():
/// Set the number of running jobs to zero
outstanding_sem(),
outstanding(0)
{
}

/// Submit a job to the threadpool
void adevs_threadpool::submit_job(job_t& job)
{
	job_cvar.lock();
	outstanding++;
	jobs.add(job);
	job_cvar.signal();
	job_cvar.unlock();
}

/// Get a job from the threadpool
adevs_threadpool::job_t adevs_threadpool::get_job()
{
	job_cvar.lock();
	while (jobs.getSize() == 0)
	{
		job_cvar.wait();
	}
	job_t result(jobs[0]);
	jobs.remove(0);
	job_cvar.unlock();
	return result;
}

/**
Assigns a thread to the job, returning true if a thread could
be found (or created) and false otherwise.
*/
bool adevs_threadpool::fork(void (*func)(void*,void*), void* arg1, void* arg2)
{
	/// Create a job to be run
	job_t job;
	job.outstanding_sem = &outstanding_sem;
	job.func = func;
	job.arg1 = arg1;
	job.arg2 = arg2;
	/// Get exclusive access to the global job list
	// If a thread is available to run the job
	if (avail_threads > 0)
	{
		submit_job(job);
		return true;
	}
	// otherwise, if a thread can be created to handle the job
	else if (total_threads < thread_limit)
	{
		/// Create the new thread
		thread* thr = new thread(run_pthread,NULL);
		submit_job(job);
		return true;
	}
	// otherwise, if no threads can be found
	return false;
}

/// Wait until all of the users jobs have terminated
void adevs_threadpool::join_all()
{
	while (outstanding > 0)
	{
		outstanding_sem.down();
		outstanding--;
	}
}

adevs_threadpool::~adevs_threadpool()
{
	join_all();
}
